# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2002-2011 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2002-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import logging, os, string, tempfile
from . import baseclasses, bbox, config, unit, trafo, pswriter

logger = logging.getLogger("pyx")

# PostScript-procedure definitions (cf. 5002.EPSF_Spec_v3.0.pdf)
# with important correction in EndEPSF:
#   end operator is missing in the spec!

_BeginEPSF = pswriter.PSdefinition("BeginEPSF", b"""{
  /b4_Inc_state save def
  /dict_count countdictstack def
  /op_count count 1 sub def
  userdict begin
  /showpage { } def
  0 setgray 0 setlinecap
  1 setlinewidth 0 setlinejoin
  10 setmiterlimit [ ] 0 setdash newpath
  /languagelevel where
  {pop languagelevel
  1 ne
    {false setstrokeadjust false setoverprint
    } if
  } if
} bind""")

_EndEPSF = pswriter.PSdefinition("EndEPSF", b"""{
  end
  count op_count sub {pop} repeat
  countdictstack dict_count sub {end} repeat
  b4_Inc_state restore
} bind""")


class linefilereader:
    """a line by line file reader

    This line by line file reader allows for '\n', '\r' and
    '\r\n' as line separation characters. Line separation
    characters are not modified (binary mode). It implements
    a readline, a read and a close method similar to a regular
    file object."""

    # note: '\n\r' is not considered to be a linebreak as its documented
    #       in the DSC spec #5001, while '\n\r' *is* a *single* linebreak
    #       according to the EPSF spec #5002

    def __init__(self, file, typicallinelen=257):
        """Opens the file filename for reading.

        typicallinelen defines the default buffer increase
        to find the next linebreak."""

        # note: The maximal line size in an EPS is 255 plus the
        #       linebreak characters. However, we also handle
        #       lines longer than that.
        self.file = file
        self.buffer = b""
        self.typicallinelen = typicallinelen

    def read(self, count=None, EOFmsg="unexpected end of file"):
        """read bytes from the file

        count is the number of bytes to be read when set. Then count
        is unset, the rest of the file is returned. EOFmsg is used
        to raise a IOError, when the end of the file is reached while
        reading count bytes or when the rest of the file is empty when
        count is unset. When EOFmsg is set to None, less than the
        requested number of bytes might be returned."""
        if count is not None:
            if count > len(self.buffer):
                self.buffer += self.file.read(count - len(self.buffer))
            if EOFmsg is not None and len(self.buffer) < count:
                raise IOError(EOFmsg)
            result = self.buffer[:count]
            self.buffer = self.buffer[count:]
            return result
        else:
            self.buffer += self.file.read()
            if EOFmsg is not None and not len(self.buffer):
                raise IOError(EOFmsg)
            result = self.buffer
            self.buffer = ""
            return result

    def readline(self, EOFmsg="unexpected end of file"):
        """reads a line from the file

        Lines are separated by '\n', '\r' or '\r\n'. The line separation
        strings are included in the return value. The last line might not
        end with an line separation string. Reading beyond the file generates
        an IOError with the EOFmsg message. When EOFmsg is None, an empty
        string is returned when reading beyond the end of the file."""
        EOF = 0
        while True:
            crpos = self.buffer.find(b"\r")
            nlpos = self.buffer.find(b"\n")
            if nlpos == -1 and (crpos == -1 or crpos == len(self.buffer) - 1) and not EOF:
                newbuffer = self.file.read(self.typicallinelen)
                if not len(newbuffer):
                    EOF = 1
                self.buffer += newbuffer
            else:
                eol = len(self.buffer)
                if not eol and EOFmsg is not None:
                    raise IOError(EOFmsg)
                if nlpos != -1:
                    eol = nlpos + 1
                if crpos != -1 and (nlpos == -1 or crpos < nlpos - 1):
                    eol = crpos + 1
                result = self.buffer[:eol]
                self.buffer = self.buffer[eol:]
                return result


def _readbbox(file):
    """returns bounding box of EPS file filename"""

    file = linefilereader(file)

    # check the %! header comment
    if not file.readline().startswith(b"%!"):
        raise IOError("file doesn't start with a '%!' header comment")

    bboxatend = 0
    # parse the header (use the first BoundingBox)
    while True:
        line = file.readline()
        if not line:
            break
        if line.startswith(b"%%BoundingBox:") and not bboxatend:
            values = line.split(b":", 1)[1].split()
            if values == ["(atend)"]:
                bboxatend = 1
            else:
                if len(values) != 4:
                    raise IOError("invalid number of bounding box values")
                return bbox.bbox_pt(*list(map(int, values)))
        elif (line.rstrip() == b"%%EndComments" or
              (len(line) >= 2 and chr(line[0]) != "%" and chr(line[1]) not in string.whitespace)):
            # implicit end of comments section
            break
    if not bboxatend:
        raise IOError("no bounding box information found")

    # parse the body
    nesting = 0 # allow for nested documents
    while True:
        line = file.readline()
        if line.startswith(b"%%BeginData:"):
            values = line.split(":", 1)[1].split()
            if len(values) > 3:
                raise IOError("invalid number of arguments")
            if len(values) == 3:
                if values[2] == b"Lines":
                    for i in range(int(values[0])):
                        file.readline()
                elif values[2] != b"Bytes":
                    raise IOError("invalid bytesorlines-value")
                else:
                    file.read(int(values[0]))
            else:
                file.read(int(values[0]))
            line = file.readline()
            # ignore tailing whitespace/newline for binary data
            if (len(values) < 3 or values[2] != "Lines") and not len(line.strip()):
                line = file.readline()
            if line.rstrip() != b"%%EndData":
                raise IOError("missing EndData")
        elif line.startswith(b"%%BeginBinary:"):
            file.read(int(line.split(":", 1)[1]))
            line = file.readline()
            # ignore tailing whitespace/newline
            if not len(line.strip()):
                line = file.readline()
            if line.rstrip() != b"%%EndBinary":
                raise IOError("missing EndBinary")
        elif line.startswith(b"%%BeginDocument:"):
            nesting += 1
        elif line.rstrip() == b"%%EndDocument":
            if nesting < 1:
                raise IOError("unmatched EndDocument")
            nesting -= 1
        elif not nesting and line.rstrip() == b"%%Trailer":
            break

    usebbox = None
    # parse the trailer (use the last BoundingBox)
    line = True
    while line:
        line = file.readline(EOFmsg=None)
        if line.startswith("%%BoundingBox:"):
            values = line.split(b":", 1)[1].split()
            if len(values) != 4:
                raise IOError("invalid number of bounding box values")
            # According to the standard, bound box values should be ints, but let's 
            # be graceful and accept floats, as well.
            usebbox = bbox.bbox_pt(*list(map(float, values)))
    if not usebbox:
        raise IOError("missing bounding box information in document trailer")
    return usebbox


class epsfile(baseclasses.canvasitem):

    """class for epsfiles"""

    def __init__(self,
                 x, y, filename,
                 width=None, height=None, scale=None, align="bl",
                 clip=1, translatebbox=1, bbox=None,
                 kpsearch=0):
        """inserts epsfile

        Object for an EPS file named filename at position (x,y). Width, height,
        scale and aligment can be adjusted by the corresponding parameters. If
        clip is set, the result gets clipped to the bbox of the EPS file. If
        translatebbox is not set, the EPS graphics is not translated to the
        corresponding origin. If bbox is not None, it overrides the bounding
        box in the epsfile itself. If kpsearch is set then filename is searched
        using the kpathsea library.
        """

        self.x_pt = unit.topt(x)
        self.y_pt = unit.topt(y)
        self.filename = filename
        self.kpsearch = kpsearch
        if bbox:
            self.mybbox = bbox
        else:
            with self.open() as epsfile:
                self.mybbox = _readbbox(epsfile)

        # determine scaling in x and y direction
        self.scalex = self.scaley = scale

        if width is not None or height is not None:
            if scale is not None:
                raise ValueError("cannot set both width and/or height and scale simultaneously")
            if height is not None:
                self.scaley = unit.topt(height)/(self.mybbox.ury_pt-self.mybbox.lly_pt)
            if width is not None:
                self.scalex = unit.topt(width)/(self.mybbox.urx_pt-self.mybbox.llx_pt)

            if self.scalex is None:
                self.scalex = self.scaley
            if self.scaley is None:
                self.scaley = self.scalex

        # set the actual width and height of the eps file (after a
        # possible scaling)
        self.width_pt  = self.mybbox.urx_pt-self.mybbox.llx_pt
        if self.scalex:
            self.width_pt *= self.scalex

        self.height_pt = self.mybbox.ury_pt-self.mybbox.lly_pt
        if self.scaley:
            self.height_pt *= self.scaley

        # take alignment into account
        self.align       = align
        if self.align[0]=="b":
            pass
        elif self.align[0]=="c":
            self.y_pt -= self.height_pt/2.0
        elif self.align[0]=="t":
            self.y_pt -= self.height_pt
        else:
            raise ValueError("vertical alignment can only be b (bottom), c (center), or t (top)")

        if self.align[1]=="l":
            pass
        elif self.align[1]=="c":
            self.x_pt -= self.width_pt/2.0
        elif self.align[1]=="r":
            self.x_pt -= self.width_pt
        else:
            raise ValueError("horizontal alignment can only be l (left), c (center), or r (right)")

        self.clip = clip
        self.translatebbox = translatebbox

        self.trafo = trafo.translate_pt(self.x_pt, self.y_pt)

        if self.scalex is not None:
            self.trafo = self.trafo * trafo.scale_pt(self.scalex, self.scaley)

        if translatebbox:
            self.trafo = self.trafo * trafo.translate_pt(-self.mybbox.llx_pt, -self.mybbox.lly_pt)

    def open(self):
        if self.kpsearch:
            return config.open(self.filename, [config.format.pict])
        else:
            return open(self.filename, "rb")

    def bbox(self):
        return self.mybbox.transformed(self.trafo)

    def processPS(self, file, writer, context, registry, bbox):
        registry.add(_BeginEPSF)
        registry.add(_EndEPSF)
        bbox += self.bbox()

        file.write("BeginEPSF\n")

        if self.clip:
            llx_pt, lly_pt, urx_pt, ury_pt = self.mybbox.transformed(self.trafo).highrestuple_pt()
            file.write("%g %g %g %g rectclip\n" % (llx_pt, lly_pt, urx_pt-llx_pt, ury_pt-lly_pt))

        self.trafo.processPS(file, writer, context, registry)

        file.write("%%%%BeginDocument: %s\n" % self.filename)

        with self.open() as epsfile:
            file.write_bytes(epsfile.read())

        file.write("%%EndDocument\n")
        file.write("EndEPSF\n")

    def processPDF(self, file, writer, context, registry, bbox):
        logger.warning("EPS file is included as a bitmap created using pipeGS")
        from pyx import bitmap, canvas
        from PIL import Image
        c = canvas.canvas()
        c.insert(self)
        i = Image.open(c.pipeGS(device="pngalpha", resolution=600))
        i.load()
        b = bitmap.bitmap_pt(self.bbox().llx_pt, self.bbox().lly_pt, i)
        # we slightly shift the bitmap to re-center it, as the bitmap might contain some additional border
        # unfortunately we need to construct another bitmap instance for that ...
        b = bitmap.bitmap_pt(self.bbox().llx_pt + 0.5*(self.bbox().width_pt()-b.bbox().width_pt()),
                             self.bbox().lly_pt + 0.5*(self.bbox().height_pt()-b.bbox().height_pt()), i)
        b.processPDF(file, writer, context, registry, bbox)



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2002-2012, 2022 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2003-2011 Michael Schindler <m-schindler@users.sourceforge.net>
# Copyright (C) 2002-2012, 2022 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import functools, logging, math
from pyx import attr, unit, text, utils
from pyx.graph.axis import painter, parter, positioner, rater, texter, tick

logger = logging.getLogger("pyx")
class _marker: pass


class axisdata:
    """axis data storage class

    Instances of this class are used to store axis data local to the
    graph. It will always contain an axispos instance provided by the
    graph during initialization."""

    def __init__(self, **kwargs):
        for key, value in list(kwargs.items()):
            setattr(self, key, value)


class _axis:
    """axis"""

    def createlinked(self, data, positioner, graphtextengine, errorname, linkpainter):
        canvas = painter.axiscanvas(self.painter, graphtextengine)
        if linkpainter is not None:
            linkpainter.paint(canvas, data, self, positioner)
        return canvas


class NoValidPartitionError(RuntimeError):

    pass


class _regularaxis(_axis):
    """base implementation a regular axis

    Regular axis have a continuous variable like linear axes,
    logarithmic axes, time axes etc."""

    def __init__(self, min=None, max=None, reverse=0, divisor=None, title=None,
                       painter=painter.regular(), texter=texter.default(), linkpainter=painter.linked(),
                       density=1, maxworse=2, manualticks=[], fallbackrange=None, **kwargs):

        split_kwargs, rest_kwargs = utils.kwsplit(kwargs, ["painter", "texter", "linkpainter"])
        if rest_kwargs:
            raise ValueError("Invalid split kwargs: %s" % rest_kwargs)

        if "painter" in split_kwargs: painter = painter(**split_kwargs["painter"])
        if "texter" in split_kwargs: texter = texter(**split_kwargs["texter"])
        if "linkpainter" in split_kwargs: linkpainter = linkpainter(**split_kwargs["linkpainter"])

        if min is not None and max is not None and min > max:
            min, max, reverse = max, min, not reverse
        self.min = min
        self.max = max
        self.reverse = reverse
        self.divisor = divisor
        self.title = title
        self.painter = painter
        self.texter = texter
        self.linkpainter = linkpainter
        self.density = density
        self.maxworse = maxworse
        self.manualticks = self.checkfraclist(manualticks)
        self.fallbackrange = fallbackrange

    def createdata(self, errorname):
        return axisdata(min=self.min, max=self.max)

    zero = 0.0

    def adjustaxis(self, data, columndata, graphtextengine, errorname):
        if self.min is None or self.max is None:
            for value in columndata:
                try:
                    value = value + self.zero
                except Exception:
                    pass
                else:
                    if self.min is None and (data.min is None or value < data.min):
                        data.min = value
                    if self.max is None and (data.max is None or value > data.max):
                        data.max = value

    def checkfraclist(self, fracs):
        "orders a list of fracs, equal entries are not allowed"
        if not len(fracs): return []
        sorted = list(fracs)
        sorted.sort()
        last = sorted[0]
        for item in sorted[1:]:
            if last == item:
                raise ValueError("duplicate entry found")
            last = item
        return sorted

    def _create(self, data, positioner, graphtextengine, parter, rater, errorname):
        errorname = " for axis %s" % errorname
        if data.min is None or data.max is None:
            raise RuntimeError("incomplete axis range%s" % errorname)
        if data.max == data.min:
            if self.fallbackrange is not None:
                try:
                    data.min, data.max = data.min - 0.5*self.fallbackrange, data.min + 0.5*self.fallbackrange
                except TypeError:
                    data.min, data.max = self.fallbackrange[0], self.fallbackrange[1]
            else:
                raise RuntimeError("zero axis range%s" % errorname)

        if self.divisor is not None:
            rational_divisor = tick.rational(self.divisor)
            convert_tick = lambda x: float(x)*self.divisor
        else:
            convert_tick = lambda x: x

        def layout(data):
            if data.ticks:
                self.adjustaxis(data, [convert_tick(data.ticks[0]), convert_tick(data.ticks[-1])], graphtextengine, errorname)
            self.texter.labels(data.ticks)
            if self.divisor:
                for t in data.ticks:
                    t *= rational_divisor
            canvas = painter.axiscanvas(self.painter, graphtextengine)
            if self.painter is not None:
                self.painter.paint(canvas, data, self, positioner)
            return canvas

        if parter is None:
            data.ticks = self.manualticks
            return layout(data)

        # a variant is a data copy with local modifications to test several partitions
        @functools.total_ordering
        class variant:
            def __init__(self, data, **kwargs):
                self.data = data
                for key, value in list(kwargs.items()):
                    setattr(self, key, value)

            def __getattr__(self, key):
                return getattr(data, key)

            def __lt__(self, other):
                # we can also sort variants by their rate
                return self.rate < other.rate

            def __eq__(self, other):
                # we can also sort variants by their rate
                return self.rate == other.rate

        # build a list of variants
        bestrate = None
        if self.divisor is not None:
            if data.min is not None:
                data_min_divided = data.min/self.divisor
            else:
                data_min_divided = None
            if data.max is not None:
                data_max_divided = data.max/self.divisor
            else:
                data_max_divided = None
            partfunctions = parter.partfunctions(data_min_divided, data_max_divided,
                                                 self.min is None, self.max is None)
        else:
            partfunctions = parter.partfunctions(data.min, data.max,
                                                 self.min is None, self.max is None)
        variants = []
        for partfunction in partfunctions:
            worse = 0
            while worse < self.maxworse:
                worse += 1
                ticks = partfunction()
                if ticks is None:
                    break
                ticks = tick.mergeticklists(self.manualticks, ticks, mergeequal=0)
                if ticks:
                    rate = rater.rateticks(self, ticks, self.density)
                    if rate is not None:
                        if self.reverse:
                            rate += rater.raterange(self.convert(data, convert_tick(ticks[0])) -
                                                    self.convert(data, convert_tick(ticks[-1])), 1)
                        else:
                            rate += rater.raterange(self.convert(data, convert_tick(ticks[-1])) -
                                                    self.convert(data, convert_tick(ticks[0])), 1)
                        if bestrate is None or rate < bestrate:
                            bestrate = rate
                            worse = 0
                        variants.append(variant(data, rate=rate, ticks=ticks))

        if not variants:
            raise RuntimeError("no axis partitioning found%s" % errorname)

        if len(variants) == 1 or self.painter is None:
            # When the painter is None, we could sort the variants here by their rating.
            # However, we didn't did this so far and there is no real reason to change that.
            data.ticks = variants[0].ticks
            return layout(data)

        # build the layout for best variants
        for variant in variants:
            variant.storedcanvas = None
        variants.sort()
        while not variants[0].storedcanvas:
            variants[0].storedcanvas = layout(variants[0])
            ratelayout = rater.ratelayout(variants[0].storedcanvas, self.density)
            if ratelayout is None:
                del variants[0]
                if not variants:
                    raise NoValidPartitionError("no valid axis partitioning found%s" % errorname)
            else:
                variants[0].rate += ratelayout
            variants.sort()
        self.adjustaxis(data, variants[0].ticks, graphtextengine, errorname)
        data.ticks = variants[0].ticks
        return variants[0].storedcanvas


class linear(_regularaxis):
    """linear axis"""

    def __init__(self, parter=parter.autolinear(), rater=rater.linear(), **kwargs):
        split_kwargs, rest_kwargs = utils.kwsplit(kwargs, ["parter", "rater"])

        _regularaxis.__init__(self, **rest_kwargs)

        if "parter" in split_kwargs: parter = parter(**split_kwargs["parter"])
        if "rater" in split_kwargs: rater = rater(**split_kwargs["rater"])

        self.parter = parter
        self.rater = rater
        self.kwargs = kwargs

    def __call__(self, **kwargs):
        return linear(**utils.merge_members_kwargs(self, [self.kwargs, kwargs], ["parter", "rater"]))

    def convert(self, data, value):
        """axis coordinates -> graph coordinates"""
        if self.reverse:
            return (data.max - float(value)) / (data.max - data.min)
        else:
            return (float(value) - data.min) / (data.max - data.min)

    def create(self, data, positioner, graphtextengine, errorname):
        return _regularaxis._create(self, data, positioner, graphtextengine, self.parter, self.rater, errorname)

lin = linear


class logarithmic(_regularaxis):
    """logarithmic axis"""

    def __init__(self, parter=parter.autologarithmic(), rater=rater.logarithmic(),
                       linearparter=parter.autolinear(extendtick=None), **kwargs):
        split_kwargs, rest_kwargs = utils.kwsplit(kwargs, ["parter", "rater", "linearparter"])

        _regularaxis.__init__(self, **rest_kwargs)

        if "parter" in split_kwargs: parter = parter(**split_kwargs["parter"])
        if "rater" in split_kwargs: rater = rater(**split_kwargs["rater"])
        if "linearparter" in split_kwargs: linearparter = linearparter(**split_kwargs["linearparter"])

        self.parter = parter
        self.rater = rater
        self.linearparter = linearparter
        self.kwargs = kwargs

    def __call__(self, **kwargs):
        return linear(**utils.merge_members_kwargs(self, [self.kwargs, kwargs], ["parter", "rater", "linearparter"]))

    def convert(self, data, value):
        """axis coordinates -> graph coordinates"""
        # TODO: store log(data.min) and log(data.max)
        if self.reverse:
            return (math.log(data.max) - math.log(float(value))) / (math.log(data.max) - math.log(data.min))
        else:
            return (math.log(float(value)) - math.log(data.min)) / (math.log(data.max) - math.log(data.min))

    def create(self, data, positioner, graphtextengine, errorname):
        try:
            return _regularaxis._create(self, data, positioner, graphtextengine, self.parter, self.rater, errorname)
        except NoValidPartitionError:
            if self.linearparter:
                logger.warning("no valid logarithmic partitioning found for axis %s, switch to linear partitioning" % errorname)
                return _regularaxis._create(self, data, positioner, graphtextengine, self.linearparter, self.rater, errorname)
            raise

log = logarithmic


class subaxispositioner(positioner._positioner):
    """a subaxis positioner"""

    def __init__(self, basepositioner, subaxis):
        self.basepositioner = basepositioner
        self.vmin = subaxis.vmin
        self.vmax = subaxis.vmax
        self.vminover = subaxis.vminover
        self.vmaxover = subaxis.vmaxover

    def vbasepath(self, v1=None, v2=None):
        if v1 is not None:
            v1 = self.vmin+v1*(self.vmax-self.vmin)
        else:
            v1 = self.vminover
        if v2 is not None:
            v2 = self.vmin+v2*(self.vmax-self.vmin)
        else:
            v2 = self.vmaxover
        return self.basepositioner.vbasepath(v1, v2)

    def vgridpath(self, v):
        return self.basepositioner.vgridpath(self.vmin+v*(self.vmax-self.vmin))

    def vtickpoint_pt(self, v, axis=None):
        return self.basepositioner.vtickpoint_pt(self.vmin+v*(self.vmax-self.vmin))

    def vtickdirection(self, v, axis=None):
        return self.basepositioner.vtickdirection(self.vmin+v*(self.vmax-self.vmin))


class bar(_axis):

    def __init__(self, subaxes=None, defaultsubaxis=linear(painter=None, linkpainter=None, parter=None),
                       dist=0.5, firstdist=None, lastdist=None, title=None, reverse=0,
                       painter=painter.bar(), linkpainter=painter.linkedbar(), **kwargs):
        split_kwargs, rest_kwargs = utils.kwsplit(kwargs, ["defaultsubaxis", "painter", "linkpainter"], allow_only_split=True)

        if "defaultsubaxis" in split_kwargs: defaultsubaxis = linear(**split_kwargs["defaultsubaxis"])
        if "painter" in split_kwargs: painter = painter.bar(**split_kwargs["painter"])
        if "linkpainter" in split_kwargs: linkpainter = painter.linkedbar(**split_kwargs["linkpainter"])

        self.subaxes = subaxes
        self.defaultsubaxis = defaultsubaxis
        self.dist = dist
        if firstdist is not None:
            self.firstdist = firstdist
        else:
            self.firstdist = 0.5 * dist
        if lastdist is not None:
            self.lastdist = lastdist
        else:
            self.lastdist = 0.5 * dist
        self.title = title
        self.reverse = reverse
        self.painter = painter
        self.linkpainter = linkpainter
        self.kwargs = kwargs

    def __call__(self, **kwargs):
        return bar(**utils.merge_members_kwargs(self, [self.kwargs, kwargs], ["defaultsubaxis", "painter", "linkpainter"]))

    def createdata(self, errorname):
        data = axisdata(size=self.firstdist+self.lastdist-self.dist, subaxes={}, names=[])
        return data

    def addsubaxis(self, data, name, subaxis, graphtextengine, errorname):
        subaxis = anchoredaxis(subaxis, graphtextengine, "%s, subaxis %s" % (errorname, name))
        subaxis.setcreatecall(lambda: None)
        subaxis.sized = hasattr(subaxis.data, "size")
        if subaxis.sized:
            data.size += subaxis.data.size
        else:
            data.size += 1
        data.size += self.dist
        data.subaxes[name] = subaxis
        if self.reverse:
            data.names.insert(0, name)
        else:
            data.names.append(name)

    def adjustaxis(self, data, columndata, graphtextengine, errorname):
        for value in columndata:

            # some checks and error messages
            try:
                len(value)
            except Exception:
                raise ValueError("tuple expected by bar axis '%s'" % errorname)
            try:
                value + ""
            except Exception:
                pass
            else:
                raise ValueError("tuple expected by bar axis '%s'" % errorname)
            assert len(value) == 2, "tuple of size two expected by bar axis '%s'" % errorname

            name = value[0]
            if name is not None and name not in data.names:
                if self.subaxes:
                    if self.subaxes[name] is not None:
                        self.addsubaxis(data, name, self.subaxes[name], graphtextengine, errorname)
                else:
                    self.addsubaxis(data, name, self.defaultsubaxis, graphtextengine, errorname)
        for name in data.names:
            subaxis = data.subaxes[name]
            if subaxis.sized:
                data.size -= subaxis.data.size
            subaxis.axis.adjustaxis(subaxis.data,
                                    [value[1] for value in columndata if value[0] == name],
                                    graphtextengine,
                                    "%s, subaxis %s" % (errorname, name))
            if subaxis.sized:
                data.size += subaxis.data.size

    def convert(self, data, value):
        if value[0] is None:
            raise ValueError
        axis = data.subaxes[value[0]]
        vmin = axis.vmin
        vmax = axis.vmax
        return axis.vmin + axis.convert(value[1]) * (axis.vmax - axis.vmin)

    def create(self, data, positioner, graphtextengine, errorname):
        canvas = painter.axiscanvas(self.painter, graphtextengine)
        v = 0
        position = self.firstdist
        for name in data.names:
            subaxis = data.subaxes[name]
            subaxis.vmin = position / float(data.size)
            if subaxis.sized:
                position += subaxis.data.size
            else:
                position += 1
            subaxis.vmax = position / float(data.size)
            position += 0.5*self.dist
            subaxis.vminover = v
            if name == data.names[-1]:
                subaxis.vmaxover = 1
            else:
                subaxis.vmaxover = position / float(data.size)
            subaxis.setpositioner(subaxispositioner(positioner, subaxis))
            subaxis.create()
            for layer, subcanvas in list(subaxis.canvas.layers.items()):
                canvas.layer(layer).insert(subcanvas)
            assert len(subaxis.canvas.layers) == len(subaxis.canvas.items)
            if canvas.extent_pt < subaxis.canvas.extent_pt:
                canvas.extent_pt = subaxis.canvas.extent_pt
            position += 0.5*self.dist
            v = subaxis.vmaxover
        if self.painter is not None:
            self.painter.paint(canvas, data, self, positioner)
        return canvas

    def createlinked(self, data, positioner, graphtextengine, errorname, linkpainter):
        canvas = painter.axiscanvas(self.painter, graphtextengine)
        for name in data.names:
            subaxis = data.subaxes[name]
            subaxis = linkedaxis(subaxis, name)
            subaxis.setpositioner(subaxispositioner(positioner, data.subaxes[name]))
            subaxis.create()
            for layer, subcanvas in list(subaxis.canvas.layers.items()):
                canvas.layer(layer).insert(subcanvas)
            assert len(subaxis.canvas.layers) == len(subaxis.canvas.items)
            if canvas.extent_pt < subaxis.canvas.extent_pt:
                canvas.extent_pt = subaxis.canvas.extent_pt
        if linkpainter is not None:
            linkpainter.paint(canvas, data, self, positioner)
        return canvas


class nestedbar(bar):

    def __init__(self, defaultsubaxis=bar(dist=0, painter=None, linkpainter=None), **kwargs):
        bar.__init__(self, defaultsubaxis=defaultsubaxis, **kwargs)


class split(bar):

    def __init__(self, defaultsubaxis=linear(),
                       firstdist=0, lastdist=0,
                       painter=painter.split(), linkpainter=painter.linkedsplit(), **kwargs):
        bar.__init__(self, defaultsubaxis=defaultsubaxis,
                           firstdist=firstdist, lastdist=lastdist,
                           painter=painter, linkpainter=linkpainter, **kwargs)


class sizedlinear(linear):

    def __init__(self, size=1, **kwargs):
        linear.__init__(self, **kwargs)
        self.size = size
        self.kwargs = kwargs

    def __call__(self, **kwargs):
        return sizedlinear(**utils.merge_members_kwargs(self, [self.kwargs, kwargs], ["size"]))

    def createdata(self, errorname):
        data = linear.createdata(self, errorname)
        data.size = self.size
        return data

sizedlin = sizedlinear


class autosizedlinear(linear):

    def __init__(self, parter=parter.autolinear(extendtick=None), **kwargs):
        split_kwargs, rest_kwargs = utils.kwsplit(kwargs, ["parter"])
        if "parter" in split_kwargs: parter = parter.autolinear(**split_kwargs["parter"])

        linear.__init__(self, parter=parter, **rest_kwargs)

        self.kwargs = kwargs

    def __call__(self, **kwargs):
        return autosizedlinear(**utils.merge_members_kwargs(self, [self.kwargs, kwargs], ["parter"]))

    def createdata(self, errorname):
        data = linear.createdata(self, errorname)
        try:
            data.size = data.max - data.min
        except Exception:
            data.size = 0
        return data

    def adjustaxis(self, data, columndata, graphtextengine, errorname):
        linear.adjustaxis(self, data, columndata, graphtextengine, errorname)
        try:
            data.size = data.max - data.min
        except Exception:
            data.size = 0

    def create(self, data, positioner, graphtextengine, errorname):
        min = data.min
        max = data.max
        canvas = linear.create(self, data, positioner, graphtextengine, errorname)
        if min != data.min or max != data.max:
            raise RuntimeError("range change during axis creation of autosized linear axis")
        return canvas

autosizedlin = autosizedlinear


class anchoredaxis:

    def __init__(self, axis, graphtextengine, errorname):
        assert not isinstance(axis, anchoredaxis), errorname
        self.axis = axis
        self.errorname = errorname
        self.graphtextengine = graphtextengine
        self.data = axis.createdata(self.errorname)
        self.canvas = None
        self.positioner = None

    def setcreatecall(self, function, *args, **kwargs):
        self._createfunction = function
        self._createargs = args
        self._createkwargs = kwargs

    def docreate(self):
        if not self.canvas:
            self._createfunction(*self._createargs, **self._createkwargs)

    def setpositioner(self, positioner):
        assert positioner is not None, self.errorname
        assert self.positioner is None, self.errorname
        self.positioner = positioner

    def convert(self, x):
        self.docreate()
        return self.axis.convert(self.data, x)

    def adjustaxis(self, columndata):
        if self.canvas is None:
            self.axis.adjustaxis(self.data, columndata, self.graphtextengine, self.errorname)
        else:
            logger.warning("ignore axis range adjustment of already created axis '%s'"  % self.errorname)

    def vbasepath(self, v1=None, v2=None):
        return self.positioner.vbasepath(v1=v1, v2=v2)

    def basepath(self, x1=None, x2=None):
        self.docreate()
        if x1 is None:
            if x2 is None:
                return self.positioner.vbasepath()
            else:
                return self.positioner.vbasepath(v2=self.axis.convert(self.data, x2))
        else:
            if x2 is None:
                return self.positioner.vbasepath(v1=self.axis.convert(self.data, x1))
            else:
                return self.positioner.vbasepath(v1=self.axis.convert(self.data, x1),
                                                 v2=self.axis.convert(self.data, x2))

    def vgridpath(self, v):
        return self.positioner.vgridpath(v)

    def gridpath(self, x):
        self.docreate()
        return self.positioner.vgridpath(self.axis.convert(self.data, x))

    def vtickpoint_pt(self, v):
        return self.positioner.vtickpoint_pt(v)

    def vtickpoint(self, v):
        return self.positioner.vtickpoint_pt(v) * unit.t_pt

    def tickpoint_pt(self, x):
        self.docreate()
        return self.positioner.vtickpoint_pt(self.axis.convert(self.data, x))

    def tickpoint(self, x):
        self.docreate()
        x_pt, y_pt = self.positioner.vtickpoint_pt(self.axis.convert(self.data, x))
        return  x_pt * unit.t_pt, y_pt * unit.t_pt

    def vtickdirection(self, v):
        return self.positioner.vtickdirection(v)

    def tickdirection(self, x):
        self.docreate()
        return self.positioner.vtickdirection(self.axis.convert(self.data, x))

    def create(self):
        if self.canvas is None:
            assert self.positioner is not None, self.errorname
            self.canvas = self.axis.create(self.data, self.positioner, self.graphtextengine, self.errorname)
        return self.canvas


class linkedaxis(anchoredaxis):

    def __init__(self, linkedaxis=None, errorname="manual-linked", painter=_marker):
        self.painter = painter
        self.linkedto = None
        self.errorname = errorname
        self.canvas = None
        self.positioner = None
        if linkedaxis:
            self.setlinkedaxis(linkedaxis)

    def setlinkedaxis(self, linkedaxis):
        assert isinstance(linkedaxis, anchoredaxis), self.errorname
        self.linkedto = linkedaxis
        self.axis = linkedaxis.axis
        self.graphtextengine = self.linkedto.graphtextengine
        self.errorname = "%s (linked to %s)" % (self.errorname, linkedaxis.errorname)
        self.data = linkedaxis.data
        if self.painter is _marker:
            self.painter = linkedaxis.axis.linkpainter

    def create(self):
        assert self.linkedto is not None, self.errorname
        assert self.positioner is not None, self.errorname
        if self.canvas is None:
            self.linkedto.docreate()
            self.canvas = self.axis.createlinked(self.data, self.positioner, self.graphtextengine, self.errorname, self.painter)
        return self.canvas


class anchoredpathaxis(anchoredaxis):
    """an anchored axis along a path"""

    def __init__(self, path, axis, **kwargs):
        anchoredaxis.__init__(self, axis, text.defaulttextengine, "pathaxis")
        self.setpositioner(positioner.pathpositioner(path, **kwargs))
        self.create()

def pathaxis(*args, **kwargs):
    """creates an axiscanvas for an axis along a path"""
    return anchoredpathaxis(*args, **kwargs).canvas




from pyx import *

c = canvas.canvas()
c.stroke(path.line(0, 0, 3, 0))
c.stroke(path.rect(0, 1, 1, 1))
c.fill(path.circle(2.5, 1.5, 0.5))
c.writeEPSfile("path")
c.writePDFfile("path")
c.writeSVGfile("path")



from pyx import *

c = canvas.canvas()

c.stroke(path.line(-5, 0, -5, 5))
c.stroke(path.line(0, 0, 0, 5))
c.stroke(path.line(5, 0, 5, 5))

c.text(-5, 5, r"boxleft", [text.halign.boxleft])
c.text(0, 5, r"boxcenter", [text.halign.boxcenter])
c.text(5, 5, r"boxright", [text.halign.boxright])

c.text(0, 4, r"boxcenter and flushleft",
       [text.parbox(3), text.halign.boxcenter, text.halign.flushleft])
c.text(0, 3, r"boxcenter and flushcenter",
       [text.parbox(3), text.halign.boxcenter, text.halign.flushcenter])
c.text(0, 2, r"boxcenter and flushright",
       [text.parbox(3), text.halign.boxcenter, text.halign.flushright])

c.text(-5, 0, r"left: boxleft and flushleft",
       [text.parbox(3), text.halign.left])
c.text(0, 0, r"center: boxcenter and flushcenter",
       [text.parbox(3), text.halign.center])
c.text(5, 0, r"right: boxright and flushright",
       [text.parbox(3), text.halign.right])

c.writeEPSfile("halign")
c.writePDFfile("halign")
c.writeSVGfile("halign")



import math
from pyx import *
from pyx.graph import axis

texter1 = axis.texter.decimal(plus="+", minus="-", equalprecision=1)
texter2 = axis.texter.rational()

class piaxis(axis.linear):

    def __init__(self, divisor=math.pi,
                 texter=axis.texter.rational(suffix=r"\pi"), **kwargs):
        axis.linear.__init__(self, divisor=divisor, texter=texter, **kwargs)

p1 = path.path(path.moveto(0, 0), path.curveto(3, 0, 1, 4, 4, 4))
p2 = p1.transformed(trafo.translate(4, 0))
p3 = p2.transformed(trafo.translate(4, 0))
p4 = p3.transformed(trafo.translate(4, 0))

c = canvas.canvas()
c.insert(axis.pathaxis(p1, axis.linear(min=0, max=1e2)))
c.insert(axis.pathaxis(p2, axis.linear(min=-0.7, max=0.4, texter=texter1)))
c.insert(axis.pathaxis(p3, axis.linear(min=-0.7, max=0.4, texter=texter2)))
c.insert(axis.pathaxis(p4, piaxis(min=0, max=2*math.pi)))
c.writeEPSfile("texter")
c.writePDFfile("texter")
c.writeSVGfile("texter")



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2002-2012, 2021 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2003-2006 Michael Schindler <m-schindler@users.sourceforge.net>
# Copyright (C) 2002-2012 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import math
from . import attr, unit

#
# base classes for stroke and fill styles
#

class style:
    pass

class strokestyle(style):
    pass

class fillstyle(style):
    pass

#
# common stroke styles
#


class linecap(attr.exclusiveattr, strokestyle):

    """linecap of paths"""

    def __init__(self, value=0):
        attr.exclusiveattr.__init__(self, linecap)
        self.value = value

    def __eq__(self, other):
        return self.value == other.value if isinstance(other, self.__class__) else NotImplemented

    def processPS(self, file, writer, context, registry):
        file.write("%d setlinecap\n" % self.value)

    def processPDF(self, file, writer, context, registry):
        file.write("%d J\n" % self.value)

    def processSVGattrs(self, attrs, writer, context, registry):
        attrs["stroke-linecap"] = {0: "butt", 1: "round", 2: "square"}[self.value]


linecap.butt = linecap(0)
linecap.round = linecap(1)
linecap.square = linecap(2)
linecap.clear = attr.clearclass(linecap)


class linejoin(attr.exclusiveattr, strokestyle):

    """linejoin of paths"""

    def __init__(self, value=0):
        attr.exclusiveattr.__init__(self, linejoin)
        self.value = value

    def __eq__(self, other):
        return self.value == other.value if isinstance(other, self.__class__) else NotImplemented

    def processPS(self, file, writer, context, registry):
        file.write("%d setlinejoin\n" % self.value)

    def processPDF(self, file, writer, context, registry):
        file.write("%d j\n" % self.value)

    def processSVGattrs(self, attrs, writer, context, registry):
        attrs["stroke-linejoin"] = {0: "miter", 1: "round", 2: "bevel"}[self.value]

linejoin.miter = linejoin(0)
linejoin.round = linejoin(1)
linejoin.bevel = linejoin(2)
linejoin.clear = attr.clearclass(linejoin)


class miterlimit(attr.exclusiveattr, strokestyle):

    """miterlimit of paths"""

    def __init__(self, value=10.0):
        attr.exclusiveattr.__init__(self, miterlimit)
        self.value = value

    def __eq__(self, other):
        return self.value == other.value if isinstance(other, self.__class__) else NotImplemented

    def processPS(self, file, writer, context, registry):
        file.write("%f setmiterlimit\n" % self.value)

    def processPDF(self, file, writer, context, registry):
        file.write("%f M\n" % self.value)

    def processSVGattrs(self, attrs, writer, context, registry):
        attrs["stroke-miterlimit"] = "%f" % self.value

miterlimit.lessthan180deg = miterlimit(1/math.sin(math.pi*180/360))
miterlimit.lessthan90deg = miterlimit(1/math.sin(math.pi*90/360))
miterlimit.lessthan60deg = miterlimit(1/math.sin(math.pi*60/360))
miterlimit.lessthan45deg = miterlimit(1/math.sin(math.pi*45/360))
miterlimit.lessthan11deg = miterlimit(10) # the default, approximately 11.4783 degress
miterlimit.clear = attr.clearclass(miterlimit)


_defaultlinewidth = 0.02 * unit.w_cm
_defaultlinewidth_pt = unit.topt(_defaultlinewidth)

class dash(attr.exclusiveattr, strokestyle):

    """dash of paths"""

    def __init__(self, pattern=[], offset=0, rellengths=1):
        """set pattern with offset.

        If rellengths is True, interpret all dash lengths relative to current linewidth.
        """
        attr.exclusiveattr.__init__(self, dash)
        self.pattern = pattern
        self.offset = offset
        self.rellengths = rellengths

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.pattern == other.pattern and self.offset == other.offset and self.rellengths == other.rellengths
        return NotImplemented

    def processPS(self, file, writer, context, registry):
        if self.rellengths:
            patternstring = " ".join(["%f" % (element * context.linewidth_pt/_defaultlinewidth_pt) for element in self.pattern])
        else:
            patternstring = " ".join(["%f" % element for element in self.pattern])
        file.write("[%s] %f setdash\n" % (patternstring, self.offset))

    def processPDF(self, file, writer, context, registry):
        if self.rellengths:
            patternstring = " ".join(["%f" % (element * context.linewidth_pt/_defaultlinewidth_pt) for element in self.pattern])
        else:
            patternstring = " ".join(["%f" % element for element in self.pattern])
        file.write("[%s] %f d\n" % (patternstring, self.offset))

    def processSVGattrs(self, attrs, writer, context, registry):
        if self.rellengths:
            patternstring = " ".join(["%f" % (element * context.linewidth_pt/_defaultlinewidth_pt) for element in self.pattern])
        else:
            patternstring = " ".join(["%f" % element for element in self.pattern])
        if patternstring:
            attrs["stroke-dasharray"] = patternstring
            attrs["stroke-dashoffset"] = "%f" % self.offset
        else:
            attrs["stroke-dasharray"] = "none"

dash.clear = attr.clearclass(dash)


class linestyle(attr.exclusiveattr, strokestyle):

    """linestyle (linecap together with dash) of paths"""

    def __init__(self, c=linecap.butt, d=dash([])):
        # XXX better, but at the moment not supported by attr.exlusiveattr would be:
        # XXX   attr.exclusiveattr.__init__(self, [linestyle, linecap, dash])
        attr.exclusiveattr.__init__(self, linestyle)
        self.c = c
        self.d = d

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.c == other.c and self.d == other.d
        return NotImplemented

    def processPS(self, file, writer, context, registry):
        self.c.processPS(file, writer, context, registry)
        self.d.processPS(file, writer, context, registry)

    def processPDF(self, file, writer, context, registry):
        self.c.processPDF(file, writer, context, registry)
        self.d.processPDF(file, writer, context, registry)

    def processSVGattrs(self, attrs, writer, context, registry):
        self.c.processSVGattrs(attrs, writer, context, registry)
        self.d.processSVGattrs(attrs, writer, context, registry)

linestyle.solid = linestyle(linecap.butt, dash([]))
linestyle.dashed = linestyle(linecap.butt, dash([2]))
linestyle.dotted = linestyle(linecap.round, dash([0, 2]))
linestyle.dashdotted = linestyle(linecap.round, dash([0, 2, 2, 2]))
linestyle.clear = attr.clearclass(linestyle)


class linewidth(attr.sortbeforeexclusiveattr, strokestyle):

    """linewidth of paths"""

    def __init__(self, width):
        attr.sortbeforeexclusiveattr.__init__(self, linewidth, [dash, linestyle])
        self.width = width

    def __eq__(self, other):
        return self.width == other.width if isinstance(other, self.__class__) else NotImplemented

    def processPS(self, file, writer, context, registry):
        context.linewidth_pt = unit.topt(self.width)
        file.write("%f setlinewidth\n" % context.linewidth_pt)

    def processPDF(self, file, writer, context, registry):
        context.linewidth_pt = unit.topt(self.width)
        file.write("%f w\n" % context.linewidth_pt)

    def processSVGattrs(self, attrs, writer, context, registry):
        context.linewidth_pt = unit.topt(self.width)
        attrs["stroke-width"] = "%f" % context.linewidth_pt

linewidth.THIN = linewidth(_defaultlinewidth/math.sqrt(32))
linewidth.THIn = linewidth(_defaultlinewidth/math.sqrt(16))
linewidth.THin = linewidth(_defaultlinewidth/math.sqrt(8))
linewidth.Thin = linewidth(_defaultlinewidth/math.sqrt(4))
linewidth.thin = linewidth(_defaultlinewidth/math.sqrt(2))
linewidth.normal = linewidth(_defaultlinewidth)
linewidth.thick = linewidth(_defaultlinewidth*math.sqrt(2))
linewidth.Thick = linewidth(_defaultlinewidth*math.sqrt(4))
linewidth.THick = linewidth(_defaultlinewidth*math.sqrt(8))
linewidth.THIck = linewidth(_defaultlinewidth*math.sqrt(16))
linewidth.THICk = linewidth(_defaultlinewidth*math.sqrt(32))
linewidth.THICK = linewidth(_defaultlinewidth*math.sqrt(64))
linewidth.clear = attr.clearclass(linewidth)


class fillrule(attr.exclusiveattr, fillstyle):

    """defines the fill rule to be used"""

    def __init__(self, even_odd):
        attr.exclusiveattr.__init__(self, fillrule)
        self.even_odd = even_odd

    def __eq__(self, other):
        return self.even_odd == other.even_odd if isinstance(other, self.__class__) else NotImplemented

    def processPS(self, file, writer, context, registry):
        context.fillrule = self.even_odd

    def processPDF(self, file, writer, context, registry):
        context.fillrule = self.even_odd

    def processSVGattrs(self, attrs, writer, context, registry):
        attrs["fill-rule"] = {0: "nonzero", 1: "evenodd"}[self.even_odd]

fillrule.nonzero_winding = fillrule(0)
fillrule.even_odd = fillrule(1)
fillrule.clear = attr.clearclass(fillrule)
